/* 
 * Description: CGI to save a paswword change request
 * to a file in a defined directory.
 * Licence:  MIT
 * Author:  Xavier Cartron prx@si3t.ch
 */


#include <ctype.h>
#include <err.h>
#include <regex.h>
#include <limits.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/syslimits.h>

#include "config.h"

#define LEN(X)	(sizeof X / sizeof X[0])


char *		get_post_field(const char *, const char *, char *);
void		http200(const char *);
void		http400(const char *);
void		http500(const char *);
int		ishex(int);
size_t		rdmstr(char *, size_t);
int		urldecode(const char *, char *);
int		strtosecret(const char *, char *);


char *
get_post_field(const char *str, const char *post, char *field)
{
	char *pos		= NULL;
	char buf[BUFSIZ] 	= {'\0'};


	char *r = strdup(post);
	char *tofree = r;
	char *tok = NULL;

	while ((tok = strsep(&r, "&")) != NULL) {
		if ((pos = strstr(tok, str)) == NULL) {
			continue;
		} else {
			if (strlcpy(buf, tok + strlen(str),
				sizeof(buf)) >= sizeof(buf))
					http500("strlcpy");
			break;
		}
	}

	if (urldecode(buf, field) < 0)
		http400("Bad request");

	free(tofree);
	return field;
}

void http200(const char *content_type)
{
	printf("%s\n", "Status: 200 OK");
	printf("Content-Type: %s\n", content_type);
	printf("\n");
}

void
http400(const char *msg)
{
	printf("Status: 400 Bad Request\n");
	printf("Content-Type: text/html; charset=utf-8\n\n");
	printf("%s\n", msg);
	err(1, "%s", msg);
}

void
http500(const char *msg)
{
	printf("Status: 500 Internal Server Error\n");
	printf("Content-Type: text/html; charset=utf-8\n\n");
	printf("%s\n", msg);
	err(1, "%s", msg);
}

/* https://rosettacode.org/wiki/URL_decoding#C */

int ishex(int x)
{
	return	(x >= '0' && x <= '9')	||
		(x >= 'a' && x <= 'f')	||
		(x >= 'A' && x <= 'F');
}

size_t
rdmstr(char *s, size_t len)
{
	/* fill s with random chars */

	size_t l=0;

	const char charset[] =
		"0123456789"
		"abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMOPQRSTUVWXYZ"
		"-_+~#()%,?;.:/!=";

	for (size_t i = 0; i < len; i++) {
		s[i] = charset[arc4random_uniform(strlen(charset))];
		l++;
	}
	s[len] = '\0';

	return l;
	
}

int
strtosecret(const char *s, char *secret)
{
	/* choose a random class of char 
	 * secret is all char of the selected class
	 * from string s
	 */

	int char_classes[] = { DIGIT, LOWER, UPPER, PUNCT };
	int char_class = char_classes[arc4random_uniform(LEN(char_classes))];

	for (size_t i=0; i < strlen(s); i++) {
		switch (char_class) {
		case DIGIT:
			if (isdigit(s[i]))
				secret[strlen(secret)] = s[i];
			break;
		case LOWER:
			if (islower(s[i]))
				secret[strlen(secret)] = s[i];
			break;
		case UPPER:
			if (isupper(s[i]))
				secret[strlen(secret)] = s[i];
			break;
		case PUNCT:
			if (ispunct(s[i]))
				secret[strlen(secret)] = s[i];
			break;
		}
	}
	
	return char_class;
}

int urldecode(const char *s, char *dec)
{
	char *o;
	const char *end = s + strlen(s);
	int c;

	for (o = dec; s <= end; o++) {
		c = *s++;
		if (c == '+') c = ' ';
		else if (c == '%' && (	!ishex(*s++)	||
					!ishex(*s++)	||
					!sscanf(s - 2, "%2x", &c)))
			return -1;

		if (dec)
			*o = c;
	}

	return o - dec;
}

int 
main(int argc, char *argv[])
{
	char rawpost[BUFSIZ]		= {'\0'};
	char user[LOGIN_NAME_MAX]	= {'\0'};
	char recovery[MAXLEN]		= {'\0'};
	char answer[MAXLEN]		= {'\0'};
	char newpass[MAXLEN]		= {'\0'};
	char anshash[_PASSWORD_LEN]	= {'\0'};
	char question[CAPTCHALEN]	= {'\0'};
	char secret[CAPTCHALEN]		= {'\0'};
	char request_file[PATH_MAX]	= {'\0'};
	FILE *f 			= NULL;
	int  secret_class;
	regex_t reg_username;


	/* only the data dir can be accessed thanks to unveil */
	/* plege, btw */

#ifdef __OpenBSD__
	/*
	if (unveil(chpw_requests_dir, "rwc") == -1)
		err(1, "unveil");
	unveil(NULL, NULL);
	if (pledge("stdio cpath wpath", NULL) == -1)
		err(1, "pledge");
	*/
#endif

	/* first, wait to avoid bruteforce overflow */
	sleep(arc4random_uniform(MAXWAIT));

	/* Check if data is POSTed in stdin */
	if (fgets(rawpost, sizeof(rawpost), stdin) == NULL)
		if (ferror(stdin))
			http400("Request too long ?");

	/* remove ending newline (for testing in cli) */
	rawpost[strcspn(rawpost, "\n")] = '\0';

	if (strlen(rawpost) == 0) {

		/* no data posted: display form:
		 *
		 * Create a random captcha
		 * hash it to send it in the form
		 */

		/* Roll as long as answer is not long enough */
		while (strlen(secret) < 6) {
			//secret[0] = '\0';
			memset(secret, '\0', sizeof(secret));
			rdmstr(question, sizeof(question));
			secret_class = strtosecret(question, secret);
		}

		/* create hash from secret */
		if (crypt_newhash(secret, "bcrypt,a", anshash,
					sizeof(anshash)) != 0)
			http500("can't generate hash");

		http200("text/html; charset=utf-8");
		printf(form,
			instructions[secret_class],
			question,
			anshash
		);

	} else {
		/* there is data in post
		 * get all fields
		 * check captcha
		 * then save request in a file
		 */

		get_post_field("user=", rawpost, user);
		get_post_field("recovery=", rawpost, recovery);
		get_post_field("newpass=", rawpost, newpass);
		get_post_field("answer=", rawpost, answer);
		get_post_field("anshash=", rawpost, anshash);

		/* check if answer match hash */
		if (crypt_checkpass(answer, anshash) != 0)
			http400(errcaptcha);

		/* check if inputs are valid */

		/* user first */
		if (regcomp(&reg_username, valid_username_regex,
				REG_EXTENDED) != 0) {
			regfree(&reg_username);
			http500("can't compile regex");
		}

		if (regexec(&reg_username, user, 0, NULL, 0) != 0) {
			regfree(&reg_username);
			http400(errusername);
		}
		regfree(&reg_username);

		/* check password length */
		if (strlen(newpass) < 15)
			http400(errshortpw);

		/* build the path to file with request */
		if ((strlcpy(request_file, chpw_requests_dir, sizeof(request_file))
			> sizeof(request_file)) ||
			(strlcat(request_file, "/", sizeof(request_file))
			> sizeof(request_file)) ||
			(strlcat(request_file, user, sizeof(request_file))
			> sizeof(request_file)))
				http500("strlcpy & strlcat");


		/* open the file and save data */
		f = fopen(request_file, "w");
		if (f == NULL)
			http500("can't open file to write");
		fprintf(f, "%s\n", recovery);
		fprintf(f, "%s\n", newpass);
		fclose(f);
		chmod(request_file,
			S_IRUSR | S_IWUSR |
			S_IRGRP | S_IWGRP |
			S_IROTH | S_IWOTH
			);

		http200("text/html; charset=utf-8");
		printf("%s", okform);
	}

	return 0;
}
